import numpy as np 
import pandas as pd 

from ..depth.depth import *

class ClusterSet:
    def __init__(self, data=list()):
        self.clusters = data
    
    def add_cluster(self, data):
        self.clusters.append(data)
    
    def get_cluster(self, position):
        return self.clusters[position]
    
    def clusters(self):
        return self.clusters
    
    def set_cluster(self, idx, data):
        self.clusters[idx] = data
        
    def __len__(self):
        return len(self.clusters)
    
    def __iter__(self):
        for df in self.clusters:
            yield df 
    
    def display(self, cluster=None):
        if cluster is not None:
            display(self.clusters[cluster])
        for c in self.clusters:
            display(c)
    
def init_clusterset(data: pd.DataFrame, k: int):
    n = data.shape[1]
    cset = ClusterSet()
    
    for i in range(k):
        sample = data.sample(n=n // k, axis=1, replace=False)
        cset.add_cluster(sample)
        
    return cset



import pandas as pd 
import numpy as np
from typing import List, Union
from scipy.special import erf, binom
from scipy.integrate import tplquad, quad
import scipy.stats as stats
from scipy.stats import norm
from scipy.special import gamma, gammaincc, factorial
from tqdm import tqdm
from numpy import exp
from scipy.special import hyp2f1, binom
from scipy.stats import poisson

from ._functional import _subsequences
from ._helper import *


__all__ = ['probabilistic_normal_depth', 'probabilistic_poisson_depth']

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Poisson
def _poisson_containment(lambda_i: float, lambda_f: float, lambda_j: float, lim: int, tol: float, quiet=False) -> float:
    '''Explicitely calculates Pr(f_i <= f <= f_j) where f_i is the Poisson distribution with mean lambda_i, f is the 
    Poisson distribution with mean lambda_f, and f_j is the Poisson distribution with lambda_j'''
    s = 0
    
    for z in range(1, lim):
        ls = sum(poisson.pmf(k=l, mu=lambda_i) for l in range(z))
        ks = sum(poisson.pmf(k=k, mu=lambda_j) for k in range(z, lim))
        c = ls * ks * poisson.pmf(k=z, mu=lambda_f)

        s += c

    return s

def _poisson_containment_simplified(lambda_i, lambda_f, lambda_j, lim=100, quiet=False) -> float:
    '''Calculates Pr(f_i <= f <= f_j) like above, but uses the simplification found by Wolfram in order to speed up computation.
    This method takes about half the time. '''
    return exp(-lambda_f) * sum(
        (np.power(lambda_f, z)*(gamma(z)-gammainc(z, lambda_j))*gammainc(1+z, lambda_i))/\
        (factorial(z)*gamma(z)*gamma(1+z)) for z in range(1, lim)
    )

def _poisson_depth(df: pd.DataFrame, curr: int, lim: int, tol: float, to_compute=None):
    n, p = df.shape
    S_nj = 0
    cols = list(df.columns)
    cols.remove(curr)
    
    subseq = _subsequences(cols, 2)
    
    for sequence in subseq:
        i, j = sequence
        S_nj += sum(
            _poisson_containment_simplified(
                df.loc[k, i], 
                df.loc[k, curr], 
                df.loc[k, j],
                lim,
                tol,) 
            for k in df.index)
        
    return S_nj

def probabilistic_poisson_depth(df: pd.DataFrame, to_compute=None, lim=1000, tol=10**-6):
    n, p = df.shape
    depths = []
    for f in tqdm(df.columns):
        depths.append(1/binom(n, 2) * _poisson_depth(df, f, lim, tol, to_compute))
    
    return pd.Series(index=df.columns, data=depths)

# # # # # # # # # # # # # # # # # # # # # # # #

# # # # # # # # # # # # # # # # # # # # # # # #
# Binomial
def _binom_containment(params: list, lim=3000) -> float:
    n_i, p_i, n, p, n_j, p_j = params
    
    s = 0
    for z in range(1, lim):
        t = (1-p)**(n-z)*p**z*(1-p_i)**n_i*(1-p_j)**(n_j-z)*p_j**z*\
             binom(n, z)*binom(n_j, z)*\
             ((1/(1-p_i))**n_i - (1-p_i)**(-1-z)*p_i**(1+z)*binom(n_i, 1+z)*\
             hyp2f1(1,1-n_i+z, 2+z, p_i/(p_i-1)))*hyp2f1(1, -n_j+z, 1+z, p_j/(p_j-1))
        
        if np.isnan(t):
            print(t)
            break
            
    return s

# # # # # # # # # # # # # # # # # # # # # # # #

# # # # # # # # # # # # # # # # # # # # # # # #
# Normal

def _normal_containment(z, parameters: list):
    mu_i, sigma_i, mu_j, sigma_j, mu, sigma = parameters
    return (normcdf(z, mu_i, sigma_i)-normcdf(z, mu, sigma)*normcdf(z, mu_j, sigma_j))*\
            norm(loc=mu, scale=sigma).pdf(z)

def _normal_depth(means, stds, curr, f):
    n = len(means)
    cols = list(range(n))
    cols.remove(curr)
    S_nj = 0
    subseq = _subsequences(cols, 2)

    for i in tqdm(range(len(subseq))):
        sequence = subseq[i]
        i, j = sequence
        
        parameters = [
            means[i], stds[i], 
            means[j], stds[j], 
            means[curr], stds[curr]
        ]
        
        integral = quad(lambda x: f(x, parameters), -np.inf, np.inf)[0]
        S_nj += integral
        
    return S_nj / binom(n, 2)

def probabilistic_normal_depth(means, stds, f):
    if len(means) != len(stds):
        raise ValueError('Error, len(means) must equal len(stds)')

    depths = []
    for k in tqdm(range(len(means))):
        mc = np.delete(means, [k])
        stdsc = np.delete(stds, [k])
        
        depths.append(    
            _normal_depth(means, stds, k, f)
        )
    
    return pd.DataFrame({
        'means': means,
        'stds': stds,
        'depths': depths
    })


import pandas as pd 
from typing import Callable, List, Union, Dict
import plotly.graph_objects as go

from .calculations._helper import *

from .calculations._functional import _functionaldepth, _samplefunctionaldepth
from .calculations._pointcloud import _pointwisedepth, _samplepointwisedepth
from .abstract import AbstractDepth

__all__ = ['FunctionalDepth', 'PointcloudDepth', 'ProbabilisticDepth']

# Private class that wraps the band depth calculation methods with some extra attributes as well
class _FunctionalDepthSeries(AbstractDepth, pd.Series):

    def __init__(self, df: pd.DataFrame, depths: pd.Series):
        super().__init__(data=depths)

        self._orig_data = df
        self._depths = depths
        self._ordered_depths = None

    def ordered(self, ascending=False) -> pd.Series:
        '''Sort the curves by their band depth, from deepest to most outlying'''
        if self._ordered_depths is None:
            self._ordered_depths =  self._depths.sort_values(ascending=ascending)
        return self._ordered_depths

    def deepest(self, n=1) -> pd.Series:
        '''Return the n deepest curves. Equivalently, return the n largest items in the depths Series'''
        if self._ordered_depths is None:
            self._ordered_depths = self._depths.sort_values(ascending=False)
        
        if n == 1:
            return pd.Series(index=[list(self._ordered_depths.index)[0]], data=[self._ordered_depths.values[0]])
        else:
            return pd.Series(index=self._ordered_depths.index[0: n], data=self._ordered_depths.values[0: n])
    
    def outlying(self, n=1) -> pd.Series:
        if self._ordered_depths is None:
            self._ordered_depths = self._depths.sort_values(ascending=False)

        if n == 1:
            return pd.Series(index=[list(self._ordered_depths.index)[-1]], data=[self._ordered_depths.values[-1]])
        else:
            return pd.Series(index=self._ordered_depths.index[-n: ], data=self._ordered_depths.values[-n: ])

    # Aliases for some abstract methods above
    def sorted(self, ascending=False):
        return self.ordered(ascending=ascending)

    def median(self):
        return self.deepest(n=1)

    def get_depths(self):
        return self._depths

    def get_data(self):
        return self._orig_data

    def depths(self):
        return self.get_depths()

class _FunctionalDepthMultivariateDataFrame(AbstractDepth, pd.DataFrame):
    # Don't copy over all the data. Numpy arrays might be large and this will add extra space complexity
    # Also, visualization for multivariate functions isn't extensive enough to justify 
    # storing all of the data
    def __init__(self, depths: pd.DataFrame):
        super().__init__(depths)
        self._depths = depths

    # Not inheritable from _FunctionalDepthSeries because depths is a DataFrame
    def ordered(self, ascending=False):
        pass

    def deepest(self, n=1):
        pass

    def outlying(self, n=1):
        pass
    
# Just extends the FunctionalDepthSeries class with extra plotting capabilities,
# Since in this case our data are real-valued functions
class _FunctionalDepthUnivariate(_FunctionalDepthSeries):
    def __init__(self, df: pd.DataFrame, depths: pd.Series):
        super().__init__(df=df, depths=depths)

    def _plot(self, deep_or_outlying: pd.Series) -> None:
        cols = self._orig_data.columns
        x = self._orig_data.index

        # We use deep_or_outlying.index to get the columns because 
        # deep_or_outlying is a Series indexed by the original columns
        
        data=[go.Scatter(x=x, y=self._orig_data.loc[:, y], mode='lines', line=dict(color='Blue', width=.5)) for y in cols]
        data.extend([go.Scatter(x=x, y=self._orig_data.loc[:, y], mode='lines', line=dict(color='Red', width=2)) for y in deep_or_outlying.index])

        fig = go.Figure(data=data)
        fig.update_layout(showlegend=False)

        fig.show()

    def plot_deepest(self, n=1) -> None:
        '''Plots all the data in blue and marks the n deepest in red'''
        self._plot(deep_or_outlying=self.deepest(n=n))

    def plot_outlying(self, n=1) -> None:
        '''Plots all the data in blue and marks the n most outlying curves in red'''
        self._plot(deep_or_outlying=self.outlying(n=n))

    # Have to redefine these because in the univariate case our samples are column based
    def drop_outlying_data(self, n=1) -> pd.DataFrame:
        return self._orig_data.drop(self.outlying(n=n).index, axis=1)
    
    def get_deep_data(self, n=1) -> pd.DataFrame:
        return self._orig_data.loc[:, self.deepest(n=n).index]

class _PointwiseDepth(_FunctionalDepthSeries):
    '''Pointwise depth calculation for Multivariate data. Calculates depth of each point with respect to the sample in R^n.'''

    def __init__(self, df: pd.DataFrame, depths: pd.Series):
        super().__init__(df=df, depths=depths)

    def plot_depths(self, invert_colors=False, marker=None) -> None:
        d = self._depths
        cols = self._orig_data.columns
        n = len(self._orig_data.columns)

        if invert_colors:
            d = 1 - d

        if marker is None:
            marker = dict(color=d, colorscale='viridis', size=10)

        if n > 3:
            self._plot_parallel_axis()
        elif n == 3:
            fig = go.Figure(data=[
                go.Scatter3d(x=self._orig_data.loc[:, cols[0]], y=self._orig_data.loc[:, cols[1]], z=self._orig_data.loc[:, cols[2]], 
                mode='markers', 
                marker=marker)
            ])

            fig.update_layout(showlegend=False)
            fig.show()

        elif n == 2:
            fig = go.Figure(data=[
                go.Scatter(x=self._orig_data.loc[:, cols[0]], y=self._orig_data.loc[:, cols[1]], 
                mode='markers', 
                marker=marker)
            ])

            fig.update_layout(showlegend=False)
            fig.show()
        else: #n==1
            pass

    def _plot_parallel_axis(self) -> None:
        pass 

    def _plot(self, deep_or_outlying: pd.Series) -> None:
        n = len(self._orig_data.columns)
        cols = self._orig_data.columns
        select = self._orig_data.loc[deep_or_outlying.index, :]
        
        if n > 3:
            self._plot_parallel_axis()
        elif n == 3:
            fig = go.Figure(data=[
                go.Scatter3d(x=self._orig_data[cols[0]], y=self._orig_data[cols[1]], z=self._orig_data[cols[2]], 
                    mode='markers', marker_color='blue', name=''),
                go.Scatter3d(x=select[cols[0]], y=select[cols[1]], z=select[cols[2]], 
                    mode='markers', 
                    marker_color='red', name='')
            ])
            
            fig.update_layout(showlegend=False)
            fig.show()

        elif n == 2: 
            fig = go.Figure(data=[
                go.Scatter(x=self._orig_data[cols[0]], y=self._orig_data[cols[1]], mode='markers', marker_color='blue', name=''),
                go.Scatter(x=select[cols[0]], y=select[cols[1]], mode='markers', 
                        marker_color='red', name='')
            ])
            
            fig.update_layout(showlegend=False)
            fig.show()
        else: # n = 1, plot number line maybe
            pass
        
    def drop_outlying_data(self, n=1) -> pd.DataFrame:
        return self._orig_data.drop(self.outlying(n=n).index, axis=0)
    
    def get_deep_data(self, n=1) -> pd.DataFrame:
        return self._orig_data.loc[self.deepest(n=n).index, :]


    def plot_deepest(self, n=1) -> None:
        self._plot(self.deepest(n=n))

    def plot_outlying(self, n=1) -> None:
        self._plot(self.outlying(n=n))

    def plot_distribution(self, invert_colors=False, marker=None) -> None:
        self.plot_depths(invert_colors, marker)

# Wraps the PointwiseDepth class in a function, because we need to compute depths before we pass down to the class
def PointcloudDepth(data: pd.DataFrame, to_compute: pd.Index=None, K=None, containment='simplex', quiet=True) -> _PointwiseDepth:
    if K is not None:
        depth = _samplepointwisedepth(data=data, to_compute=to_compute, K=K, containment=containment)
    else:
        depth = _pointwisedepth(data=data, to_compute=to_compute, containment=containment)
    
    return _PointwiseDepth(df=data, depths=depth)

# Wraps FunctionalDepth classes in a function, because we need to compute depths before we pass down to the class
def FunctionalDepth(data: List[pd.DataFrame], to_compute=None, K=None, J=2, 
containment='r2', relax=False, deep_check=False, quiet=True) -> Union[_FunctionalDepthSeries, _FunctionalDepthUnivariate, _FunctionalDepthMultivariateDataFrame]:   

    # Compute band depth completely or sample band depth
    if K is not None:
        depth = _samplefunctionaldepth(
            data=data,
            to_compute=to_compute, 
            K=K, 
            J=J, 
            containment=containment, 
            relax=relax, 
            deep_check=deep_check,
            quiet=quiet
        )
    else:
        depth = _functionaldepth(
            data=data, 
            to_compute=to_compute, 
            J=J, 
            containment=containment, 
            relax=relax, 
            deep_check=deep_check,
            quiet=quiet
        )

    # Return the appropriate class
    if isinstance(depth, pd.DataFrame): 
        return _FunctionalDepthMultivariateDataFrame(depths=depth)
    elif len(data) == 1: # Univariate case (by assumption)
        return _FunctionalDepthUnivariate(df=data[0], depths=depth)
    else: # Multivariate case
        return _FunctionalDepthSeries(df=data[0], depths=depth)


import pandas as pd 
import numpy as np
from typing import Union, List

def generate_noisy_univariate(data: Union[list, np.array]=None, n: int=20, columns=None, index=None) -> pd.DataFrame:
    """
    Generate n univariate functions that are equal to the given data plus some random pertubations. 
    Should be used for testing / understanding other methods in this library.

    Parameters:
    -----------
    data: list or np.array
        1d list of numbers to generate noisy data from. 
    n: (default=20)
        Number of noisy functions to generate.
    columns: (default=None)
        Names of columns. 
    index: (default=None)
        Index to use.

    Returns:    
    ---------
    pd.DataFrame: n x p DataFrame of p real valued functions observed at n discrete time points. (So each column is a function)

    """

    df = pd.DataFrame()

    # If no starter function is provided, just generate some random one over [0,1]
    if data is None:
        data = np.random.rand(n)

    # Generate perturbed
    for k in range(n):
        r = np.random.rand()
        df[k] = np.multiply(data, r)

    if index is not None:
        df.index = index 
    if columns is not None:
        df.columns = columns

    return df

def generate_noisy_multivariate(
    data: pd.DataFrame=None, 
    num_curves: int=5, 
    n: int=10, 
    d: int=3, 
    columns=None, 
    index=None
) -> List[pd.DataFrame]:
    """
    Generate num_curves noisy multivariate functions with d features observed at n time points. 
    Should be used for testing / understanding other methods in this library.

    Parameters:
    -----------
    data: list or np.array
        1d list of numbers to generate noisy data from. 
    num_curves: (default=5)
        Number of multivariate functions to generate.
    n: (default=10)
        Number of timepoints.
    d: (default=3)
        Number of features (columns) our multivariate functions. This is the dimension of the image. 
    columns: (default=None)
        Names of columns. 
    index: (default=None)
        Index to use.

    Returns:    
    ---------
    List[pd.DataFrame]: A list of num_curves multivariate functions (DataFrames)

    """
    fs = []
    
    if data is None:
        data = np.random.rand(n, d)
        
    for k in range(num_curves):
        r = np.random.rand()
        fs.append(pd.DataFrame(data) * r)

    if index is not None:
        for df in fs:
            df.index = index

    if columns is not None:
        for df in fs:
            df.columns = columns
    
    return fs

def generate_noisy_pointcloud(
    n: int=50, 
    d: int=2, 
    columns=None, 
    index=None
) -> pd.DataFrame:
    """
    Generate n d-dimensional points from the normal distribution over [0,1]

    Parameters:
    -----------
    n: (default=20)
        Number of points to generate.
    d: (default=2)
        Dimension to draw points from.
    columns: (default=None)
        Names of columns. 
    index: (default=None)
        Index to use.

    """

    df = pd.DataFrame(np.random.rand(n, d))

    if columns is not None:
        df.columns = columns
    
    if index is not None:
        df.index = index
    
    return df


from abc import ABC, abstractmethod

class AbstractDepth(ABC):
    '''Abstract class for all Functional Depth classes. This should be used as the superclass in any future work.'''

    # Abstract methods
    # These must be implemented by subclasses
    @abstractmethod
    def ordered(self, ascending=False):
        raise NotImplementedError

    @abstractmethod
    def deepest(self, n=1):
        raise NotImplementedError

    @abstractmethod
    def outlying(self, n=1):
        raise NotImplementedError




import pandas as pd 
import numpy as np 
from typing import List, Union

from ..depth.depth import FunctionalDepth, PointcloudDepth
from ..depth.abstract import AbstractDepth

# Helper class 
class FunctionalHomogeneity:
    def __init__(self, F, G, method='p1', K=None, J=None, containment='r2', relax=False, deep_check=False, quiet=False):
        self._orig_F = F 
        self._orig_G = G 
        self._hom = _functionalhomogeneity(
            F=F,
            G=G,
            K=K,
            J=J,
            containment=containment,
            method=method,
            relax=relax,
            deep_check=deep_check,
            quiet=quiet
        )
    
    def __str__(self):
        return self._hom 
    
    def homogeneity(self):
        return self._hom 

    def __str__(self):
        return str(self.homogeneity())
    
    def __repr__(self):
        return str(self.homogeneity())

class PointcloudHomogeneity:
    def __init__(self, F, G, method='p1', K=None, J=None, containment='simplex', relax=False, deep_check=False):
        self._orig_F = F
        self._orig_G = G
        
        self._F_depths, self._G_depths, self._hom  = _pointcloudhomogeneity(
            F=F,
            G=G,
            K=K,
            containment=containment,
            method=method,
        )

    def F_depths(self):
        return self._F_depths
    
    def G_depths(self):
        return self._G_depths

    def homogeneity(self):
        return self._hom 
    
    def __str__(self):
        return str(self.homogeneity())
    
    def __repr__(self):
        return str(self.homogeneity())    

def _functionalhomogeneity(
    F: List[pd.DataFrame], 
    G: List[pd.DataFrame], 
    K=None, 
    J=2, 
    containment='r2', 
    method='p1', 
    relax=False,
    deep_check=False,
    quiet=False
):
    _handle_errors(F, G, method) 

    # Compute depths of G, needed in either case
    G_depths = FunctionalDepth(
        data=G,
        K=K,
        J=J,
        containment=containment,
        relax=relax,
        deep_check=deep_check,
        quiet=quiet,
    )
    
    # Univariate case
    if len(F) == 1:
        F = F[0]
        G = G[0]
        
        # Get deepest function in G
        G_deepest = G_depths.get_deep_data(n=1)

        # Shitty hacky fix, will change this. Sorry if someone is reading this in the future and I didn't lol
        if 'g_deepest' in F.columns:
            raise ValueError('Cannot have column named g_deepest in F')

        # Append this to F and calculate it's depth with respect to the other samples in F
        F.loc[:, 'g_deepest'] = G_deepest
        
        G_deep_in_F = FunctionalDepth(
            [F],
            to_compute=['g_deepest'],
            K=K,
            J=J,
            containment=containment,
            relax=relax,
            deep_check=deep_check,
            quiet=quiet,
        ).ordered().loc['g_deepest']
        
        F = F.drop('g_deepest', axis=1)

        if method == 'p1':
            return G_deep_in_F 
        elif method == 'p2':
            F_depths = FunctionalDepth([F], K=K, J=J, containment=containment, relax=relax, deep_check=deep_check, quiet=quiet)
            return 1 - np.abs(G_deep_in_F - F_depths.median().iloc[0])
        elif method == 'p3':
            t = []

            # Get the deepest curve of G in F
            for col in G.columns:
                F.loc[:, col] = G.loc[:, col]
                t.append(FunctionalDepth([F], to_compute=[col],  K=K, J=J, containment=containment, relax=relax, deep_check=deep_check).loc[col])
                F = F.drop(col, axis=1)

            # Sort depths of G in F
            depths_G_in_F = pd.Series(index=list(G.columns), data=t).sort_values(ascending=False)

            return depths_G_in_F.iloc[0] / G_depths.median().iloc[0]
        elif method == 'p4':
            pass
            # t1 = np.abs(self.PointcloudHomogeneity(F, G, K, J, containment, 'p3', relax, deepcheck) - self.PointcloudHomogeneity(F, F, K, J, containment, 'p1', relax, deepcheck))
            # t2 = np.abs(self.PointcloudHomogeneity(F, G, K, J, containment, 'p3', relax, deepcheck) - self.PointcloudHomogeneity(G, G, K, J, containment, 'p1', relax, deepcheck))
            # return t1 * t2
        else:
            raise ValueError(f'{method} is not a valid depth method for the given data. Use one of [\'p1\', \'p2\', \'p3\', \'p4\']')
    else:
        G_deepest = G[G_depths.index[0]] # Get deepest dataframe in G

        F.append(G_deepest)
        G_deep_in_F = FunctionalDepth(F, to_compute=[len(F) - 1], K=K, J=J, containment=containment, relax=relax, deep_check=deep_check).ordered().iloc[0]
        F.pop(-1) # Remove G after we added it 

        if method == 'p1':
            return G_deep_in_F / G_depths.median().iloc[0]
        elif method == 'p2':
            F_depths = FunctionalDepth(F, K, J, containment, relax, deep_check)
            return 1 - np.abs(G_deep_in_F - F_depths.median().iloc[0])
        elif method == 'p3':
            pass
        else:
            raise ValueError(f'{method} is not a valid depth method for the given data. Use one of [\'p1\', \'p2\', \'p3\', \'p4\']')

def _pointcloudhomogeneity(
    F: pd.DataFrame, 
    G: pd.DataFrame, 
    K=None, 
    containment='simplex', 
    method='p1', 
):
    _handle_errors(F, G, method) 

    G_depths = PointcloudDepth(data=G, K=K, containment=containment)
    F_depths = PointcloudDepth(data=F, K=K, containment=containment)
    hom = 0

    # Get deepest function in G
    G_deepest = G_depths.get_deep_data(n=1)
    G_deepest.index = ['g_deepest']

    # Append this to F and calculate it's depth with respect to the other samples in F
    F = F.append(G_deepest)
    G_deep_in_F = PointcloudDepth(F, to_compute=['g_deepest'], K=K, containment=containment).ordered().loc['g_deepest']
    F = F.drop('g_deepest', axis=0)

    if method == 'p1':
        hom =  G_deep_in_F / F_depths.median().iloc[0] # Normalized value
    elif method == 'p2':
        hom = 1 - np.abs(G_deep_in_F - F_depths.median().iloc[0])
    elif method == 'p3':
        t = []

        # Get the deepest curve of G in F
        for point in G.index:
            F.loc[point, :] = G.loc[point, :]
            t.append(PointcloudDepth(F, to_compute=[point],  K=K, containment=containment).loc[point])
            F = F.drop(point, axis=0)

        # Sort depths of G in F
        depths_G_in_F = pd.Series(index=list(G.index), data=t).sort_values(ascending=False)

        hom = depths_G_in_F.iloc[0] / G_depths.median().iloc[0]
    elif method == 'p4':
        t1 = np.abs(_pointcloudhomogeneity(F, G, K, containment, 'p3') - _pointcloudhomogeneity(F, F, K, containment, 'p1'))
        t2 = np.abs(_pointcloudhomogeneity(F, G, K, containment, 'p3') - _pointcloudhomogeneity(G, G, K, containment, 'p1'))
        hom = t1 * t2
    else:
        raise ValueError(f'{method} is not a valid depth method for the given data. Use one of [\'p1\', \'p2\', \'p3\', \'p4\']')
    
    return F_depths, G_depths, hom

def _handle_errors(F: List[pd.DataFrame], G: List[pd.DataFrame], method='p1'):
    if len(F) != len(G):
        raise ValueError('F and G must have data of the same length')
    
    if len(F) == 1: # Univariate error handling
        if F[0].shape[0] != G[0].shape[0]:
            raise ValueError('Univariate data must have same number of time indices to check containment.')
    else:
        pass




import pandas as pd 
import numpy as np

from statdepth import FunctionalDepth, PointcloudDepth
from statdepth.testing import generate_noisy_univariate, generate_noisy_multivariate, generate_noisy_pointcloud 

df = generate_noisy_univariate()
bd = FunctionalDepth([df], K=2)

print(f'univariate median is {bd.median()}')

data = generate_noisy_multivariate(num_curves=8)
bd = FunctionalDepth(data, containment='simplex', quiet=False)

print(f'multivariate depth is {bd.ordered()}')

df = generate_noisy_pointcloud()
bd = PointcloudDepth(df, K=10, quiet=False)
print(f'pointcloud median is {bd.median()}')




import numpy as np
import pandas as pd 
from typing import Union, List
from scipy.special import binom 
from scipy.spatial import ConvexHull
from tqdm import tqdm

from ._containment import _is_in_simplex
from ._helper import *

__all__ = ['_pointwisedepth', '_samplepointwisedepth']

def _pointwisedepth(
    data: pd.DataFrame, 
    to_compute: Union[list, pd.Index]=None, 
    containment='simplex', 
    quiet=True
) -> pd.Series:
    """
    Compute pointwise depth for n points in R^p, where data is an nxp matrix of points. If points is not None,
    only compute depth for the given points (should be a subset of data.index)
    
    Parameters:
    ----------
    data: pd.DataFrame
        n x d DataFrame, where we have n points in d dimensional space.
    points: list, pd.Index
        The particular points (indices) we would like to calculate band curve for. If None, we calculate depth for all points
    containment: str
        Definition of containment

    Returns:
    ----------
    pd.Series: Depth values for the given points with respect to the data. Index of Series are indices of points in the original data, and the values are the depths

    """
    n, d = data.shape
    depths = []

    if to_compute is None:
        to_compute = data.index 

    if containment == 'simplex':
        for time in tqdm(to_compute, disable=quiet):
            S_nj = 0
            
            point = data.loc[time, :]
            
            subseq = _subsequences(list(data.drop(time, axis=0).index), d + 1)

            for seq in subseq:
                S_nj += _is_in_simplex(simplex_points=
                        np.array(data.loc[seq, :]), point=np.array(point))
                
            depths.append(S_nj / binom(n, d + 1))
    elif containment == 'l1':
        return _L1_depth(data=data, to_compute=to_compute)
    elif containment == 'mahalanobis':
        return _mahalanobis_depth(data=data, to_compute=to_compute)
    elif containment == 'oja':
        return _oja_depth(data=data, to_compute=to_compute)
    else: # Probably will be more in the future 
        raise ValueError(f'{containment} is not a valid containment measure. ')

    return pd.Series(index=to_compute, data=depths)

def _samplepointwisedepth(
    data: pd.DataFrame, 
    to_compute: pd.Index=None, 
    K=2, 
    containment='simplex', 
    quiet=True
) -> pd.Series:
    """
    Compute sample pointwise depth for n points in R^p, where data is an nxp matrix of points. If points is not None,
    only compute depth for the given points (should be a subset of data.index)
    
    Parameters:
    ----------
    data: pd.DataFrame
        n x d DataFrame, where we have n points in d dimensional space.
    points: list, pd.Index
        The particular points (indices) we would like to calculate band curve for. If None, we calculate depth for all points.
    K=2:
        Number of blocks to compute sample depth with. 
    containment: str
        Definition of containment.

    Returns:
    ----------
    pd.Series: Depth values for the given points with respect to the data. Index of Series are indices of points in the original data, and the values are the depths

    """

    # If K=1, don't bother splitting the data. Just return regular depth. 
    if K == 1:
        return _pointwisedepth(data=data, to_compute=to_compute, containment=containment)

    n, d = data.shape 
    depths = []

    if to_compute is None:
        to_compute = data.index 

    # K blocks of points (indices)
    ss = n // K 

    # Compute sample depth of each point, should be containment agnostic
    # Since the computation is being done in _pointwisedepth, which will call the appropriate depth measure
    for time in tqdm(to_compute, disable=quiet):
        cd = []
        for _ in tqdm(range(ss), disable=quiet):
            sdata = data.sample(n=ss, axis=0)
            
            # If our current datapoint isnt in the sampled data, just append it since we need to sample it 
            if not time in sdata.index:
                sdata = sdata.append(data.loc[time, :])
                
            cd.append(_pointwisedepth(data=sdata, to_compute=[time], containment=containment))
        depths.append(np.mean(cd))
    
    return pd.Series(index=to_compute, data=depths)

def _L1_depth(data: pd.DataFrame, to_compute: pd.Index=None):
    """
    Computes L1 data depth of the given points. 
    """
    n, d = data.shape
    depths = []
    idx = list(data.index)

    if to_compute is None:
        to_compute = idx
    
    for point in to_compute:
        sum_e = 0
        vec = data.loc[point, :]

        # Ugly code, but faster than 
        # Computing data.drop(point).index each time point
        cidx = idx[:]
        cidx.remove(point)

        for other in cidx:
            sum_e += (data.loc[other, :] - vec) / np.linalg.norm(vec - data.loc[other, :])
            
        depths.append(np.linalg.norm(sum_e) / n)
        
    return pd.Series(index=to_compute, data=1-np.array(depths))

def _mahalanobis_depth(data: pd.DataFrame, to_compute=None):
    """Mahalanobis depth for n points in R^n. Matrix must be square since we need to multiply the inv(covariance(data)).x for each point x."""
    n, p = data.shape

    # I think? We have to multiply and nxn matrix by our point in R^p
    if n != p:
        raise ValueError('Mahalanobis depth requires equal number of dimensions and datapoints.')
        
    mu = data.mean()
    inv_cov = np.linalg.inv(np.cov(data, rowvar=True))
    idx = data.index
    
    if to_compute is not None:
        idx = to_compute
        
    depths = []
    
    for point in idx:
        x = data.loc[point, :]
        S_x = np.dot(inv_cov, x)
        
        depths.append(np.dot((x-mu).T, S_x))
    return pd.Series(index=idx, data=depths)

def _oja_depth(data: pd.DataFrame, to_compute: list=None) -> pd.Series:
    """Oja depth for n multivariate samples in R^p. Data should be an n x p matrix."""
    n, d = data.shape
    idx = data.index
    depths = []
    
    if to_compute is not None:
        idx = to_compute
    
    try:
        vol_conv_p = ConvexHull(data).volume
    except:
        raise DepthDegeneracy('Too many collinear points to compute depth of convex hull spanned by data. Try another depth method or remove collinearities.')
        
    for point in idx:
        ci = list(idx)
        ci.remove(point)
        subseq = _subsequences(ci, d)

        cd = 0
        for seq in subseq:
            try:
                hull = ConvexHull(data.loc[seq, :].append(data.loc[point, :]))
                cd += hull.volume
            except:
                raise DepthDegeneracy(f'Too many collinear points for data sampled at point {point}, will get a low depth score. Continuing...')

        depths.append(cd / vol_conv_p)

    return pd.Series(index=to_compute, data=depths)


from typing import Callable, Union, List
from inspect import signature
import warnings 

import numpy as np
import scipy as sp 
import pandas as pd

from numba import jit
from scipy.special import comb, binom
from scipy.optimize import linprog

from ._helper import *
# This is not a solution I like, but I don't want to spam the user with
# warnings when simplex containment is used, because linprog() is very whiny
np.testing.suppress_warnings()

warnings.filterwarnings("ignore")

def _is_valid_containment(containment: Callable[[pd.DataFrame, Union[pd.Series, pd.DataFrame], bool], float]) -> None: 
    """
    Checks if the given function is a valid definition for containment. Used when user passes a custom containment function.
    
    Parameters:
    ----------

    containment: Function (Callable) to check validity of 

    Returns:
    ----------
    Boolean indicating the validity of the passed function
    """

    # If we got here in _select_containment then the passed string is invalid
    if isinstance(containment, str):
        raise ValueError(f'containment argument \'{containment}\' is invalid. Use one of [\'r2\', \'r2_enum\', \'simplex \'] or a pass a custom containment function.')

    sig = signature(containment)
    params = sig.parameters

    if len(params) != 3:
        raise ValueError('Custom containment method has incorrect number of parameters. Expected 3, recieved {}'.format(len(params))) 

    return containment

def _r2_containment(data: pd.DataFrame, curve: pd.Series, relax: bool) -> float:
    """
    Produces \lambda_r with the given input data, using the standard ordering on R as the definition for containment.
    Parameters:
    ----------

    data: list 
        DataFrame of real-valued functions that define our band in R^2 (columns are time intervals, rows are functions)
    curve: pd.Series
        Function to check containment on 
    relax: bool
        If False, we use the strict definition of containment. If True, we consider the proportion of time the curve is in the band

    Returns:
    ----------
    float: If relax=False, then 0 if the function is not contained in the curve, 1 if it is. If relax=True, then we consider the proportion of time the curve is in the band, so we will return a number between 0 and 1. 
    """

    containment = 0
    
    y_range = []
    
    # Grab the mins/maxs across all rows (functions at each time index)
    mins = data.min(axis=1)
    maxs = data.max(axis=1)
    
    # Generate intervals in R over each time index
    intervals = [[i, j] for i, j in zip(mins, maxs)]
    
    # Check if each value in the curve is contained within the band 
    for index, val in enumerate(curve):
        if intervals[index][0] <= val <= intervals[index][1]:
            containment += 1
    
    # If relax=True, then we return the proportion of points in the band, else, Python integer division will round down to 0 unless all points are contained in the band (strict containment)
    return containment / len(curve) if relax else containment // len(curve)


def _r2_enum_containment(data: List[pd.DataFrame], curve: pd.DataFrame, relax: bool) -> float:
    """
    Implements the r2_enum definition of containment, where we treat each component in the vector valued function as a real valued function, 
    and calculate containment for each one. If all the components are contained in the curved defined by that componenent, then we say the function is contained.
    
    Parameters:
    ----------

    data: list
        Array of multivariate functions, where each item in array is an observation (columns are features, rows are time indices)
    curve: pd.DataFrame 
        Multivariate function to check containment of 
    relax: bool
        If False, we use the strict definition of containment. If True, we consider the proportion of time the curve is in the band

    Returns:
    ----------
    float: If relax=False, then 0 if the function is not contained in the curve, 1 if it is. If relax=True, then we consider the proportion of time the curve is in the band, so we will return a number between 0 and 1. 
    """

    raise NotImplementedError

def _simplex_containment(data: List[pd.DataFrame], curve: pd.DataFrame, relax: bool) -> float:
    """
    Implements the simplex definition of containment for multivariate functions in R^n
        
    Parameters:
    ----------

    data: list
        Array of multivariate functions, where each item in array is an observation (columns are features, rows are time indices)
    curve: pd.DataFrame 
        Multivariate function to check containment of 
    relax: bool
        If False, we use the strict definition of containment. If True, we consider the proportion of time the curve is in the band

    Returns:
    ----------
    float: If relax=False, then 0 if the function is not contained in the curve, 1 if it is. If relax=True, then we consider the proportion of time the curve is in the band, so we will return a number between 0 and 1. 
    """

    n = len(data)
    l, _ = data[0].shape
    
    containment = 0

    # For each time index, check containment 
    for idx in curve.index:
        containment += _is_in_simplex(simplex_points=np.array([df.loc[idx, :] for df in data]), 
                                point=np.array(curve.loc[idx, :]))
    
    # If relaxation, return proportion of containment, else do integer divion so that we 
    # only get 1 if all rows are contained
    return containment / l if relax else containment // l

def _is_in_simplex(simplex_points: pd.DataFrame, point: pd.Series) -> bool:
    """
    Checks if the d dimensional point is in the simplex formed by d + 1 simplex_points (geometric degeneracy allowed)
    
    Parameters:
    ----------
    
    simplex_points: pd.DataFrame
        List of n-dimensional points (rows) to form the simplex. 
    point: pd.Series
        n-dimensional point to check containment of 

    Returns:
    ----------
    bool: True if point is contained in the simplex, False otherwise
    """
    # Check if the vector x (point) can be written as a 
    # convex combination of x_1,...,x_n (simplex_points), i.e.
    # x = a_1x_1+...+a_nx_n such that a_i >=0, a_1+...+a_n = 1.

    # If this is possible, then the point is in the convex hull formed by those points.
    # In this case, the convex hull is formed with d+1 points so it is a simplex. 

    n_points = len(simplex_points)
    n_dim = len(point)
    
    c = np.zeros(n_points)
    A = np.r_[simplex_points.T, np.ones((1, n_points))]
    b = np.r_[point, np.ones(1)]
    
    # This barely ever errors, but when it does the problem is infeasible so let's assume
    # There is no containment
    try:
        lp = linprog(c, A_eq=A, b_eq=b)
    except:
        warnings.warn('Simplex computation failed. Assuming False')
        return False
    
    return lp.success

def _select_containment(containment: Union[str, Callable]) -> Callable:
    """
    Helper function to select definition of containment
    
    Parameters:
    ----------

    containment: Union[str, Callable]
        Containment string or function, used to select a built in containment method or handle a custom containment

    Returns:
    ----------
    Callable: Returns the containment function, or raises an error 

    """
    
    # Select our containment definition if it is in our pre-defined list
    if containment == 'r2':
        return _r2_containment
    elif containment == 'r2_enum':
        return _r2_enum_containment
    elif containment == 'simplex':
        return _simplex_containment
    else:
        # Otherwise check validity of the containment Callable
        return _is_valid_containment(containment=containment)



#!/usr/bin/env python

"""Tests for `statdepth` package."""

import unittest
import pandas as pd 
import numpy as np

from statdepth import *
from statdepth.testing import *
from statdepth.homogeneity import * 

class TestStatdepth(unittest.TestCase):
    """Tests for `statdepth` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_functional(self):
        """Test something."""
        df = generate_noisy_univariate()
        bd = FunctionalDepth([df], containment='r2')

        self.assertIsInstance(bd, pd.Series)
        self.assertIsInstance(bd.ordered(), pd.Series)
        self.assertIsInstance(bd.median(), pd.Series)
        self.assertIsInstance(bd.deepest(n=2), pd.Series)
        self.assertIsInstance(bd.outlying(n=2), pd.Series)

        self.assertIsInstance(FunctionalDepth([df], K=5, containment='r2'), pd.Series)

    def test_pointcloud_l1(self):
        df = generate_noisy_pointcloud(n=10, d=2)
        bd = PointcloudDepth(df, containment='l1')
        self.assertIsInstance(bd, pd.Series)
        self.assertIsInstance(bd.ordered(), pd.Series)
        self.assertIsInstance(bd.median(), pd.Series)
        self.assertIsInstance(bd.deepest(n=2), pd.Series)
        self.assertIsInstance(bd.outlying(n=2), pd.Series)


        self.assertIsInstance(PointcloudDepth(df, K=2, containment='l1'), pd.Series)
    def test_pointcloud_simplex(self):
        df = generate_noisy_pointcloud(n=10, d=2)
        bd = PointcloudDepth(df, containment='simplex')
        self.assertIsInstance(bd, pd.Series)
        self.assertIsInstance(bd.ordered(), pd.Series)
        self.assertIsInstance(bd.median(), pd.Series)
        self.assertIsInstance(bd.deepest(n=2), pd.Series)
        self.assertIsInstance(bd.outlying(n=2), pd.Series)

        self.assertIsInstance(PointcloudDepth(df, K=2, containment='simplex'), pd.Series)

    def test_pointcloud_oja(self):

        df = generate_noisy_pointcloud(n=20, d=2)
        bd = PointcloudDepth(df, containment='oja')
        self.assertIsInstance(bd, pd.Series)
        self.assertIsInstance(bd.ordered(), pd.Series)
        self.assertIsInstance(bd.median(), pd.Series)
        self.assertIsInstance(bd.deepest(n=2), pd.Series)
        self.assertIsInstance(bd.outlying(n=2), pd.Series)

        self.assertIsInstance(PointcloudDepth(df, K=2, containment='oja'), pd.Series)

    def test_multivariate(self):
        data = generate_noisy_multivariate()
        bd = FunctionalDepth(data, containment='simplex')
        self.assertIsInstance(bd, pd.Series)
        self.assertIsInstance(bd.ordered(), pd.Series)

if __name__ == "__main__":
    unittest.main()


import pandas as pd 
import numpy as np 
from typing import List, Union, Callable
from itertools import combinations
from scipy.special import binom
from scipy.integrate import quad
import numpy as np
import pandas as pd 
import scipy.stats as stats
from tqdm import tqdm

from numpy import exp
from scipy.special import erf
from scipy.integrate import quad
from scipy.stats import norm
from scipy.special import gamma, gammaincc, factorial

# Custom error class for anytime there is going to be some degeneracy with depth calculation (i.e. degenerate simplices)
class DepthDegeneracy(Exception):
    pass

def _subsequences(s: list, l: int) -> list:
    """
    Returns a list of all possible subsequences of length l from the given input list

    Parameters:
    ----------
    s: list
        List to enumerate
    l: int
        Length of subsequences to compute

    Returns:
    ----------
    list: List of subsequences
    """

    return list(set(combinations(s, l)))


def _handle_depth_errors(
    data: List[pd.DataFrame], 
    J: int, 
    containment: Union[Callable, str], 
    relax: bool, 
    deep_check: bool
) -> None:
    """
    Handles errors in band depth methods.

    Parameters:
    ----------
    data: list
        Functions to calculate band depth from
    J: int
        Parameter J for computing band depth
    containment:
        Definition of containment, either a string or bool

    Returns:
    ----------
    None: Nothing is returned, but exceptions are raised if needed
    """
    
    # Type checking for all variables
    if not isinstance(data, list):
        raise ValueError('data must be passed as a list.')

    if not isinstance(J, int):
        raise ValueError('J must be an integer.')

    if not (isinstance(containment, str) or isinstance(containment, Callable)):
        raise ValueError('containment must be of type str or Callable.')

    if not isinstance(deep_check, bool):
        raise ValueError('deep_check must be of type bool.')

    if not isinstance(relax, bool):
        raise ValueError('relax must be of type bool')
    
    # J = 0,1 doesn't make sense
    if J < 2:
        raise ValueError('Parameter J must be greater than or equal to 2.')

    # Make sure a non-empty list is passed
    if len(data) == 0:
        raise ValueError('No data passed.')

    # Make sure J < len(data) in the univariate and multivariate case
    if len(data) == 1 and J >= len(data[0]) or len(data) > 1 and J >= len(data):
        raise ValueError('Parameter J must be less than the number of observations.')
    
    if len(data) > 1 and containment == 'r2':
        raise ValueError('containment argument \'r2\' is invalid for multivariate data. Use one of [\'r2_enum\', \'simplex \'] or a passed containment method. ')

    # If there is not at least d + 2 functions for our d dimensional data, then for each function
    # We won't have d + 1 vertices to construct a simplex, which means every simplex will be at least one dimensional degenerate
    # Therefore we say depth is not well defined and error
    if isinstance(data, list) and len(data) < data[0].shape[1] + 2 and containment == 'simplex':
        raise DepthDegeneracy(f'Error: Need at least {len(data)} functions to form non-degenerate simplices in {data[0].shape + 2} dimensional space. Only have {len(data)}.')

    if deep_check:
        # Check dtypes of all columns over all DataFrames. 
        # Optional because this might be computationally expensive for very large datasets. 
        indices = []
        for df in data:
            indices.append(df.index)
            df = df.infer_objects()
            for col in df:
                if not np.issubdtype(df[col].dtype, np.number):
                    raise ValueError('DataFrame must only contain numeric dtypes.')
        # Check that all indices are the same.
        if not all([all(indices[0] == i) for i in indices]):
            raise ValueError('DataFrames indices must be the same')

def gammainc(a, x):
    return gamma(a) * gammaincc(a, x)

def normcdf(x, mu, sigma):
    return norm(loc=mu, scale=sigma).cdf(x)




#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    "plotly",
    "pandas",
    "numpy",
    "scipy",
    "numba"
]

setup_requirements = [
    "plotly",
    "pandas",
    "numpy",
    "scipy",
    "numba"
]

test_requirements = [ ]

setup(
    author="Julian Lehrer",
    author_email='jmlehrer@ucsc.edu',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="A library for computing statistical band depth for multivariate time series data",
    install_requires=[
    "plotly",
    "pandas",
    "numpy",
    "scipy",
    "numba"],
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='statdepth',
    name='statdepth',
    packages=find_packages(exclude=['tests']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/braingeneers/functional_depth_methods',
    version='0.7.8',
    zip_safe=False,
)



import random 
import warnings
from itertools import combinations
from typing import Callable, List, Union
from tqdm import tqdm

import numpy as np
import scipy as sp 
import pandas as pd

from numba import jit
from scipy.special import comb, binom

from ._containment import _r2_containment, _r2_enum_containment, _simplex_containment, _select_containment, _is_valid_containment, _is_in_simplex

from ._helper import *

__all__ = ['_functionaldepth', '_samplefunctionaldepth']

def _functionaldepth(
    data: List[pd.DataFrame], 
    to_compute: Union[list, pd.Index]=None, 
    J=2, 
    containment='r2', 
    relax=False, 
    deep_check=False,
    quiet=True
) -> Union[pd.Series, pd.DataFrame]:
    """
    Calculate the band depth for a set of functional curves.

    Band depth is an order statistic for functional data (see `fboxplot`), with
    a higher band depth indicating larger "centrality".  In analog to scalar
    data, the functional curve with highest band depth is called the median
    curve, and the band made up from the first N/2 of N curves is the 50%
    central region.

    Parameters:
    ----------
    data : list of DataFrames
        Functions to calculate band depth from
    to_compute: list of DataFrame, DataFrame, or pd.Index
        Data to compute depths of. If None, compute depth of all samples. 
    J: int (default=2)
        J parameter in the band depth calculation. J=3 can be computationally expensive for large datasets, and also does not have a closed form solution. 
    containment: Callable or string (default='r2')
        Defines what containment means for the dataset. For functions from R-->R, we use the standard ordering on R. For higher dimensional spaces, we implement a simplex method. 
        A full list can be found in the README, as well as instructions on passing a custom definition for containment.  
    relax: bool
        If True, use a strict definition of containment, else use containment defined by the proportion of time the curve is in the band. 
    deep_check: bool (default=False)
        If True, perform a more extensive error checking routine. Optional because it can be computationally expensive for large datasets. 

    Returns:
    ----------
    pd.Series, pd.DataFrame: Depth values for each function.
    """
    # Handle common errors
    _handle_depth_errors(data=data, J=J, containment=containment, relax=relax, deep_check=deep_check)

    # Select containment definition
    cdef = _select_containment(containment=containment)
    # If only one item in the list, it is the real-valued case (by assumption)
    if len(data) == 1:
        if containment == 'simplex':
            cdef = _r2_containment

            
        band_depths = []
        df = data[0]
        cols = df.columns

        if to_compute is not None:
            cols = to_compute
        
        # Calculate band depth for each sample (column)
        for col in tqdm(cols, disable=quiet):
            band_depths.append(_univariate_band_depth(data=df, curve=col, relax=relax, containment=cdef, J=J))

        # Return a series indexed by our samples
        depths = pd.Series(index=cols, data=band_depths)
    else: # Multivariate case
        if containment == 'simplex':
            depths = []

            # Get 'index' of functions, which are just their indices in the list
            f = [i for i in range(len(data))]

            if to_compute is not None:
                f = to_compute

            data_to_compute = [data[i] for i in f]
            
            # Compute band depth for each function (DataFrame)
            for cdf in tqdm(data_to_compute, disable=quiet):
                cdata = [df for df in data if df is not cdf]
                depths.append(_simplex_depth(data=cdata, curve=cdf, J=J, relax=relax))
            depths = pd.Series(index=f, data=depths)
    
    return depths

def _samplefunctionaldepth(
    data: List[pd.DataFrame], 
    K: int, 
    to_compute: Union[list, pd.Index]=None, 
    J=2, 
    containment='r2', 
    relax=False, 
    deep_check=False,
    quiet=True
) -> Union[pd.Series, pd.DataFrame]:

    """
    Calculate the sample band depth for a set of functional curves.

    This is done by
    1. Splitting the data of n curves into K blocks of size ~n/K
    2. Computing band depth with respect to each block
    3. Returning the average of these

    For K << n, this should approximate the band depth well. 

    Parameters:
    ----------
    data: list of DataFrames
        Functions to calculate band depth from
    to_compute: list of DataFrame, DataFrame, or pd.Index
        Data to compute depths of. If None, compute depth of all samples. 
    K: int 
        Computes band depth by averaging band depth across K blocks of our original curves 
    J: int (default=2)
        J parameter in the band depth calculation. J=3 can be computationally expensive for large datasets, and also does not have a closed form solution. 
    containment: Callable or string (default='r2')
        Defines what containment means for the dataset.  
    relax: bool
        If True, use a strict definition of containment, else use containment defined by the proportion of time the curve is in the band. 
    deep_check: bool (default=False)
        If True, perform a more extensive error checking routine. Optional because it can be computationally expensive for large datasets. 

    Returns:
    ----------
    pd.Series: Depth values for each function.
    """
    # This has a much lower asymtotic complexity, but a much much higher space complexity,
    # Since to sample from n functions (one for each curve, without replacement), we require copying the data
    # n - 1 times
    
    samples = []
    depths = []

    # Handle common errros
    _handle_depth_errors(data=data, J=J, containment=containment, relax=relax, deep_check=deep_check)

    cdef = _select_containment(containment=containment)
    
    # Univariate case
    if len(data) == 1:
        df = data[0]
        cols = df.columns

        if to_compute is not None:
            cols = to_compute

        orig = df.loc[:, cols]
        ss = df.shape[1] // K

        if ss == 0:
            raise DepthDegeneracy(f'Block size {K} is too large, not enough functions to sample.')
        
        if containment == 'simplex':
            cdef = _r2_containment

        # Iterate over curves (columns)
        for col in tqdm(orig.columns, disable=quiet):
            depths = []

            # For each curve, compute sample band depth using K blocks of size ~len(df)/K
            for _ in range(K):
                t = df.sample(n=ss, axis=1)
                df = df.drop(t.columns, axis=1)
                t.loc[:, col] = orig.loc[:, col]
                depths.append(_univariate_band_depth(data=t, curve=col, relax=relax, containment=cdef, J=J))
            
            # Collect these estimates
            samples.append(np.mean(depths))
            df = orig.copy()

        # Average them and return a series
        samples = pd.Series(index=df.columns, data=samples)
    else:
        # TODO: Multivariate case: partition list of shuffled DataFrames, compute band depth w.r.t those, average
        shuffled = data.copy()
        random.shuffle(shuffled)
        ss = len(data) // K 

        for _ in tqdm(range(K), disable=quiet):
            pass
        
    return samples
        
def _univariate_band_depth(
    data: pd.DataFrame, 
    curve: Union[str, int], 
    relax: bool, 
    containment: Callable, 
    J=2
) -> float:
    """
    Calculates each band depth for a given curve in the dataset. Meant for J > 2, as J=2 has a closed solution. This function is wrapped in banddepth()
    
    Parameters:
    ----------
    data: pd.DataFrame
        An n x p matrix where our rows come from R. Each observation should define a curve, in the functional sense. 
    curve: int
        The particular function we would like to calculate band curve for. Given as a column of our original DataFrame. 
    relax: bool
        If True, use a strict definition of containment, else use containment defined by the proportion of time the curve is in the band. 
    containment: Callable or str
        Function that defines containment for the particular data. For example, in R^2 this would be a discrete subset 
    J=2: int
        Parameter J in band depth. Defaulted to 2. 

    Returns:
    ----------
    float: Depth value for the given function with respect to the data passed 

    """
    
    # Initialize band depth, n (number of curves)
    band_depth = 0
    n = data.shape[1]
    
    # get curve series 
    curvedata = data.loc[:, curve]

    # Drop the curve we're calculating band depth for
    data = data.drop(curve, axis=1)

    # Compute band depth
    for j in range(2, J + 1):
        
        S_nj = 0

        # Get a list of all possible subsequences of samples (cols)
        subseq = _subsequences(list(data.columns), j)

        # Iterate over all subsequences
        for sequence in subseq:
            # Grab data for the current subsequence
            subseq_df = data.loc[:, sequence]

            # Check containment for the data given the current subsequence
            S_nj += containment(data=subseq_df, curve=curvedata, relax=relax)

        band_depth += S_nj / binom(n, j)
    
    return band_depth

def _simplex_depth(data: List[pd.DataFrame], curve: pd.DataFrame, J=2, relax=False):
    """
    Calculates simplex depth of the given curve with respect to the given data

    Parameters:
    ----------
    data: list
        List of n x d matrices, where each matrix is a function observated at n discrete time points with d features
    curve: pd.DataFrame
        The particular function we would like to calculate band curve for
    J=2: int
        Parameter J in band depth. Defaulted to 2. 
    relax: bool
        If True, use a strict definition of containment, else use containment defined by the proportion of time the curve is in the band. 

    Returns:
    ----------
    float: Depth value for the given function with respect to the data passed 

    """
    l, d = data[0].shape
    n = len(data)
    depth = 0
    
    subseq = _subsequences([i for i in range(n)], d + 1)

    for seq in subseq:
        cdata = [data[i] for i in seq]
        depth += _simplex_containment(data=cdata, curve=curve, relax=relax)
    return depth / binom(n, d + 1)
